/*
	minimal IF-SEGA driver
	Copyright 1999-2000 (C) ITO, Takayuki. All rights reserved.
*/

#include <drivers/PCI.h>
#include <drivers/ISA.h>
#include <drivers/config_manager.h>
#include <drivers/KernelExport.h>
#include <drivers/Drivers.h>
#include <support/Errors.h>
#include <support/Debug.h>
#include <stdlib.h>
#include "joystick_driver.h"
#include "saturn.h"

#define	DRIVER_NAME	"ifsega"
#define	NUM_PORTS	2

/* gameport driver cookie (per open) */

typedef struct
{
	int						port;
	bool					enhanced;
	/* below are valid under enhanced mode */
	saturn_cookie			*cookie2;	/* joystick module cookie */
	joystick_module			*joy_mod;
	joystick_module_info	joy_mod_info;
} ifsega_cookie;
	
_EXPORT int32	api_version = B_CUR_DRIVER_API_VERSION;

/* NB global variables are valid only while driver is loaded */

#define	AREA_NAME	"IF-SEGA"
static volatile uint32	*vaddr = NULL;	/* PCI */
static uint32	baseport = 0;	/* ISA */

/* I don't know how to use generic_gameport_module... */

#ifdef	USE_GENERIC_GAMEPORT_MODULE
static void *storage;
static char gameport_name [] = "generic/gameport/v1";
static generic_gameport_module *gameport;
#endif

/* -----
	null-terminated array of device names supported by this driver
----- */

static const char *my_device_name[] = {
	"joystick/" DRIVER_NAME "/0",
	"joystick/" DRIVER_NAME "/1",
	NULL
};

/* ----------
	init_hardware - called once the first time the driver is loaded
----- */
_EXPORT status_t init_hardware (void)
{
	status_t err = ENODEV;
	pci_module_info *pci;
	pci_info info;
	int i;

	/* probe PCI bus */

	if (get_module (B_PCI_MODULE_NAME, (module_info **)&pci))
		return ENOSYS;
	for (i = 0; (*pci->get_nth_pci_info)(i, &info) == B_OK; i++)
	{
		/* look for NEC PowerVR */
		if (info.vendor_id == 0x1033 && info.device_id == 0x0046)
		{
			err = B_OK;
			break;
		}
	}
	put_module (B_PCI_MODULE_NAME);

	if (err == B_OK)
	{
		/* IF-SEGA/PCI found */
		uint32	physical_address;
		area_id	area;

		physical_address = info.u.h0.base_registers [0];
		dprintf (DRIVER_NAME ": PCI found at 0x%x\n", physical_address);
		area = find_area (AREA_NAME);
		if (area < 0)
			/* device is mapped into physical memory space */
			area = map_physical_memory 
				(AREA_NAME, (void *) physical_address, B_PAGE_SIZE,
				B_ANY_KERNEL_ADDRESS, B_WRITE_AREA | B_READ_AREA, 
				(void **) &vaddr);
		if (area < 0)
			return area;

		return B_OK;
	}

	/* FIXME: should probe ISA PnP */
	return B_OK;
}

/* ----------
	init_driver - optional function - called every time the driver
	is loaded.
----- */
_EXPORT status_t init_driver (void)
{
	area_id	area;
	area_info	ainfo;

	dprintf (DRIVER_NAME ": init_driver() " __DATE__ " " __TIME__ "\n");
	if ((area = find_area (AREA_NAME)) >= 0)
	{
		/* IF-SEGA/PCI */
		get_area_info (area, &ainfo);
		vaddr = (uint32 *)((uint8 *)ainfo.address + 0x58);
	}
	else
	{
		/* IF-SEGA/ISA */
		baseport = 0x1000;		/* FIXME */
		vaddr = NULL;
	}

#ifdef	USE_GENERIC_GAMEPORT_MODULE
	if (get_module (gameport_name, (module_info **)&gameport))
	{
		dprintf (DRIVER_NAME ": failed to get gameport module\n");
		return ENOSYS;
	}
	(*gameport->create_device)(0, &storage);
#endif

	return B_OK;
}


/* ----------
	uninit_driver - optional function - called every time the driver
	is unloaded
----- */
_EXPORT void uninit_driver (void)
{
	dprintf (DRIVER_NAME ": uninit_driver()\n");
#ifdef	USE_GENERIC_GAMEPORT_MODULE
	(*gameport->delete_device)(storage);
	put_module (gameport_name);
#endif
}

/*
	IF-SEGA specific subroutines
*/

#define	CTRL_REG	0
#define	DATA_REG	1

static uint8 my_read_reg
	(int port,
	int reg)
{
	int idx;
	uint8 value;

	ASSERT (0 <= port && port < NUM_PORTS);
	ASSERT (reg == CTRL_REG || reg == DATA_REG);

	idx = ((port & 1) << 1) | (reg & 1);
	if (vaddr == NULL)
		/* IF-SEGA/ISA */
		value = read_io_8 (baseport + (idx << 1) + 1);
	else
	{
		/* IF-SEGA2/PCI */
		uint32 data;
		data = 0xff003000 | (idx << 8);
		*vaddr = data | 0xC000;
		*vaddr = data | 0x8000;
		data = *vaddr;
		*vaddr = data | 0xC000;
		value = data;
	}
	return value;
}

static void my_write_reg
	(int port,
	int reg,
	uint8 value)
{
	int idx;

	ASSERT (0 <= port && port < NUM_PORTS);
	ASSERT (reg == CTRL_REG || reg == DATA_REG);

	idx = ((port & 1) << 1) | (reg & 1);
	if (vaddr == NULL)
		/* IF-SEGA/ISA */
		write_io_8 (baseport + (idx << 1) + 1, value);
	else
	{
		/* IF-SEGA2/PCI */
		uint32	data;

		data = 0xffff3000 | (idx << 8) | value;
		*vaddr = data | 0xC000;
		*vaddr = data | 0x4000;
		*vaddr = data | 0xC000;
	}
}

/* ----------
	my_device_open - handle open() calls
----- */

static status_t my_device_open
	(const char *name,
	uint32 flags,
	ifsega_cookie **out_cookie)
{
	ifsega_cookie *cookie;
	int port;
	const char *s;

	ASSERT (name != NULL);
	ASSERT (out_cookie != NULL);
	dprintf (DRIVER_NAME ": open(%s)\n", name);

	/* check device name */

	for (port = 0; (s = my_device_name [port]) != NULL; port++)
		if (strcmp (name, s) == 0)
			break;
	if (s == NULL)
	{
		dprintf (DRIVER_NAME ": no such device\n");
		return ENODEV;
	}
	
	/* FIXME: check the port to see if joystick is actually attached */

	/* initialize a cookie */

	cookie = malloc (sizeof (ifsega_cookie));
	if (cookie == NULL)
		return ENOMEM;

	cookie->port = port;
	cookie->enhanced = false;
	cookie->cookie2 = NULL;
	*out_cookie = cookie;

#ifdef	USE_GENERIC_GAMEPORT_MODULE
	(*gameport->open_hook)(storage, flags, cookie);
#endif
	return B_OK;
}

/* ----------
	my_device_read - handle read() calls
----- */

static status_t my_device_read
	(ifsega_cookie *cookie,
	off_t position,
	void *buf,
	size_t *num_bytes)
{
	int port;
	uint32 buttons, hat;
	status_t err = B_ERROR;

	ASSERT (cookie != NULL);
	ASSERT (buf != NULL);
	ASSERT (num_bytes != NULL);

	port = cookie->port;

	/* read std pad */
	my_write_reg (port, CTRL_REG, 0x1f);

	my_write_reg (port, DATA_REG, 0xe0);
	buttons = my_read_reg (port, DATA_REG) & 15;
	my_write_reg (port, DATA_REG, 0x80);
	buttons = (buttons << 4) | (my_read_reg (port, DATA_REG) & 15);
	my_write_reg (port, DATA_REG, 0xc0);
	buttons = (buttons << 4) | (my_read_reg (port, DATA_REG) & 15);
	/* L---RXYZSACB */

	my_write_reg (port, DATA_REG, 0xa0);
	hat = my_read_reg (port, DATA_REG) & 15;

	if (cookie->enhanced && *num_bytes >= sizeof (extended_joystick))
	{
		/* enhanced mode: call joystick module */
		extended_joystick joy_data;

		cookie->cookie2->buttons = buttons;
		cookie->cookie2->hat = hat;
		(*cookie->joy_mod->read)(cookie->cookie2, port,
			&joy_data, sizeof (extended_joystick));
		memcpy (buf, &joy_data, sizeof (extended_joystick));
		*num_bytes = sizeof (extended_joystick);
		err = B_OK;
	}
	if (!cookie->enhanced && *num_bytes >= sizeof (joystick))
	{
		/* standard mode: do it myself */
		static const int ampl [4] = { 0, 32767, -32768, 0 };
		joystick joy_data;

		joy_data.timestamp = system_time ();
		joy_data.horizontal = ampl [hat >> 2];
		joy_data.vertical   = ampl [hat & 3];
		/* false if pressed */
		joy_data.button1 = (buttons & 4) ? true : false;
		joy_data.button2 = (buttons & 1) ? true : false;
		memcpy (buf, &joy_data, sizeof (joystick));
		*num_bytes = sizeof (joystick);
		err = B_OK;
	}

#ifdef	USE_GENERIC_GAMEPORT_MODULE
	(*gameport->read_hook)(cookie, position, buf, num_bytes);
#endif
	return err;
}


/* ----------
	my_device_write - handle write() calls
----- */

static status_t my_device_write
	(ifsega_cookie *cookie,
	off_t position,
	const void *buf,
	size_t *num_bytes)
{
	status_t err = B_ERROR;

	ASSERT (cookie != NULL);
	ASSERT (buf != NULL);
	ASSERT (num_bytes != NULL);

	if (cookie->enhanced && *num_bytes >= sizeof (extended_joystick))
	{
		/* enhanced mode: call joystick module */
		extended_joystick joy_data;

		memcpy (&joy_data, buf, sizeof (extended_joystick));
		(*cookie->joy_mod->force)
			(cookie->cookie2, cookie->port, (bigtime_t) 0, 
			&joy_data, sizeof (extended_joystick));
		*num_bytes = sizeof (extended_joystick);
		err = B_OK;
	}
#ifdef	USE_GENERIC_GAMEPORT_MODULE
	(*gameport->write_hook)(cookie, position, buf, num_bytes);
#endif
	return err;
}

/* ----------
	my_device_control - handle ioctl calls
----- */

static status_t my_device_control
	(ifsega_cookie *cookie,
	uint32 op,
	void *arg,
	size_t len)
{
	status_t err = B_ERROR;

	ASSERT (cookie != NULL);

	dprintf (DRIVER_NAME ": ioctl(%u)\n", op);
	switch (op)
	{
	case B_JOYSTICK_SET_DEVICE_MODULE:
	/* called by BJoystick, Joysticks pref */
	/* if this ioctl returns successfully, 
		Joysticks preference assumes that 
		the joystick product may be connected */
		ASSERT (arg != NULL);
		if (cookie->enhanced)
			break;
		if (strcmp (((joystick_module_info *)arg)->module_name, 
			MODULE_PATH) == 0)		/* FIXME */
		{
			cookie->joy_mod_info = *(joystick_module_info *)arg;
			/* obtain joystick module */
			if (get_module (cookie->joy_mod_info.module_name, 
				(module_info **) &cookie->joy_mod) == B_OK)
			{
				dprintf (DRIVER_NAME ": ioctl(B_JOYSTICK_SET_DEVICE_MODULE): get_module() OK\n");
				/* enter enhanced mode */
				cookie->enhanced = true;
				(*cookie->joy_mod->configure)
					(cookie->port, &cookie->joy_mod_info, 
					sizeof (joystick_module_info), &cookie->cookie2);
				err = B_OK;
			}
		}
		break;

	case B_JOYSTICK_GET_DEVICE_MODULE:
		ASSERT (arg != NULL);
		if (cookie->enhanced)
		{
			*(joystick_module_info *)arg = cookie->joy_mod_info;
			err = B_OK;
		}
		break;

	case B_JOYSTICK_GET_SPEED_COMPENSATION:
	case B_JOYSTICK_SET_SPEED_COMPENSATION:
	case B_JOYSTICK_GET_MAX_LATENCY:
	case B_JOYSTICK_SET_MAX_LATENCY:
	case B_JOYSTICK_SET_RAW_MODE:
	default:
		break;
	}

#ifdef	USE_GENERIC_GAMEPORT_MODULE
	(*gameport->control_hook)(cookie, op, arg, len);
#endif
	return err;
}


/* ----------
	my_device_close - handle close() calls
----- */

static status_t my_device_close
	(ifsega_cookie *cookie)
{
	ASSERT (cookie != NULL);
	dprintf (DRIVER_NAME ": close()\n");
#ifdef	USE_GENERIC_GAMEPORT_MODULE
	(*gameport->close_hook)(cookie);
#endif
	return B_OK;
}


/* -----
	my_device_free - called after the last device is closed, and after
	all i/o is complete.
----- */
static status_t my_device_free
	(ifsega_cookie *cookie)
{
	ASSERT (cookie != NULL);
	if (cookie->enhanced)
	{
		(*cookie->joy_mod->crumble)(cookie->cookie2, cookie->port);
		put_module (cookie->joy_mod_info.module_name);
	}
#ifdef	USE_GENERIC_GAMEPORT_MODULE
	(*gameport->free_hook)(cookie);
#endif
	free (cookie);
	return B_OK;
}


/* -----
	function pointers for the device hooks entry points
----- */

static device_hooks my_device_hooks = {
	my_device_open,
	my_device_close,
	my_device_free,
	my_device_control,
	my_device_read,
	my_device_write,
	NULL, NULL, NULL, NULL
};

/* ----------
	publish_devices - return a null-terminated array of devices
	supported by this driver.
----- */

_EXPORT const char **publish_devices (void)
{
	return my_device_name;
}

/* ----------
	find_device - return ptr to device hooks structure for a
	given device name
----- */

_EXPORT device_hooks *find_device
	(const char *name)
{
	ASSERT (name != NULL);
	return &my_device_hooks;
}

